/*
 * Copyright 2008 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 ** IMPORTANT NOTES:
 **
 ** * A destroyClassAd + crash can result in lost information: History
 **   message may not be sent before the crash and on startup
 **   destroyClassAd will be called but ignored because
 **   processDirtyJobs has not yet had a chance to create a
 **   JobObject. This could be fixed if newClassAd and setAttribute
 **   responded immediately, which requires a way to map key->mgmtId
 **   that does not depend on the job ad itself or a duplicate ad
 **   structure
 **
 ** ++ This could be handled if we are willing to create a new
 **    JobObject within the destroyClassAd method. Or if the whole
 **    mechanism of maintaining a dirty list was abandoned.
 **/

#define _CONDOR_ALLOW_OPEN
#include "condor_common.h"

#include "MgmtScheddPlugin.h"

#include "JobObject.h"
#include "SchedulerObject.h"
#include "SubmitterObject.h"

#include "SubmitterUtils.h"
#include "JobUtils.h"

#include "condor_qmgr.h"
#include "condor_config.h"


	// Any key that begins with the '0' char is either the
 	// header or a cluster, i.e. not a job
#define IS_JOB(key) ((key) && '0' != (key)[0])


using namespace std;
using namespace qpid::management;
using namespace qpid::framing;
using namespace qmf::mrg;
using namespace qmf::mrg::grid;
using namespace mrg::grid;


bool g_publish_jobs = true;

void
MgmtScheddPlugin::earlyInitialize()
{
	char *host;
	int port;
	char *tmp;
	string storefile;

		// Since this plugin is registered with multiple
		// PluginManagers it may be initialized more than once,
		// and we don't want that
	static bool skip = false;
	if (skip) return; skip = true;

	g_publish_jobs = param_boolean("QMF_PUBLISH_JOBS", g_publish_jobs);

	singleton = new ManagementAgent::Singleton();

	submitterAds = new SubmitterHashTable(512, &hashFuncMyString);

	ManagementAgent *agent = singleton->getInstance();

	Job::registerSelf(agent);
	qmf::mrg::grid::Scheduler::registerSelf(agent);
	Submitter::registerSelf(agent);

	port = param_integer("QMF_BROKER_PORT", 5672);
	if (NULL == (host = param("QMF_BROKER_HOST"))) {
		host = strdup("localhost");
	}

	tmp = param("QMF_STOREFILE");
	if (NULL == tmp) {
		storefile = ".schedd_storefile";
	} else {
		storefile = tmp;
		free(tmp); tmp = NULL;
	}

	agent->init(string(host), port,
				param_integer("QMF_UPDATE_INTERVAL", 10),
				true,
				storefile);

	free(host);

	scheduler = NULL;

	jobTable = new HashTable<PROC_ID, JobObject *>(10240, hashFuncPROC_ID);
	dirtyJobs = new DirtyJobsType();

	isHandlerRegistered = false;

	ReliSock *sock = new ReliSock;
	if (!sock) {
		EXCEPT("Failed to allocate Mgmt socket");
	}
	if (!sock->assign(agent->getSignalFd())) {
		EXCEPT("Failed to bind Mgmt socket");
	}
	int index;
	if (-1 == (index =
			   daemonCore->Register_Socket((Stream *) sock,
										   "Mgmt Method Socket",
										   (SocketHandlercpp)
										   &MgmtScheddPlugin::HandleMgmtSocket,
										   "Handler for Mgmt Methods.",
										   this))) {
		EXCEPT("Failed to register Mgmt socket");
	}
}


void
MgmtScheddPlugin::initialize() { };


void
MgmtScheddPlugin::shutdown()
{
	if (!param_boolean("QMF_DELETE_ON_SHUTDOWN", false)) {
		return;
	}

	dprintf(D_FULLDEBUG, "MgmtScheddPlugin: shutting down...\n");
	if (singleton) {
		delete singleton;
		singleton = NULL;
	}
}


void
MgmtScheddPlugin::update(int cmd, const ClassAd *ad)
{
	MyString hashKey;
	SubmitterObject *submitterObject;

	if (!scheduler) {
		scheduler = new SchedulerObject(singleton->getInstance());
	}

	switch (cmd) {
	case UPDATE_SCHEDD_AD:
		dprintf(D_FULLDEBUG, "Received UPDATE_SCHEDD_AD\n");
		scheduler->update(*ad);

		break;
	case UPDATE_SUBMITTOR_AD:
		dprintf(D_FULLDEBUG, "Received UPDATE_SUBMITTOR_AD\n");
		if (!GetSubmitterNameFromAd((ClassAd &)*ad, hashKey)) {
			dprintf(D_FULLDEBUG, "Could not make hashkey -- ignoring ad\n");

			return;
		}

		if (!GetSubmitter(hashKey, submitterObject)) {
			EXCEPT("Failed to get submitter: %s", hashKey.Value());
		}

		submitterObject->update(*ad);

		break;
	default:
		dprintf(D_FULLDEBUG, "Unsupported command: %s\n",
				getCollectorCommandString(cmd));
	}
}


void
MgmtScheddPlugin::archive(const ClassAd */*ad*/) { };


void
MgmtScheddPlugin::newClassAd(const char */*key*/) { };


void
MgmtScheddPlugin::setAttribute(const char *key,
							   const char */*name*/,
							   const char */*value*/)
{
//	dprintf(D_FULLDEBUG, "setAttribute: %s[%s] = %s\n", key, name, value);

	if (!g_publish_jobs) return;

	markDirty(key);
}


void
MgmtScheddPlugin::destroyClassAd(const char *key)
{
	FieldTable map;
	PROC_ID id;
	JobObject *job;

//	dprintf(D_FULLDEBUG, "destroyClassAd: %s\n", key);

	if (!g_publish_jobs) return;

	if (!IS_JOB(key)) {
		return;
	}

	id = getProcByString(key);
	if (id.cluster < 0 || id.proc < 0) {
		dprintf(D_FULLDEBUG, "Failed to parse key: %s - skipping\n", key);
		return; // This could result in something lost in history
	}

		// Cancel possible processing of deleted job
	dirtyJobs->erase(string(key));
	dprintf(D_FULLDEBUG, "Canceled processing of %s\n", key);

		// Lookup the mgmt Job associated with the key
	if (-1 == jobTable->lookup(id, job)) {
			// If it is not present there is not much we can do,
			// this is bad
		dprintf(D_ALWAYS,
				"NOTICE: job %s does not exist, data may be lost\n", key);
		return; // This could result in something lost in history

				// This can happen if a job is removed from the queue,
				// the schedd is ctrl-c'd and restarted
	}

	if (!PopulateFieldTableFromJobAd(id.cluster, id.proc, map)) {
		dprintf(D_ALWAYS, "ERROR: Failed to retrieve data for job %s\n", key);
	}

	((Job *) job->GetManagementObject())->set_Ad(map);

	delete job;
}


void
MgmtScheddPlugin::deleteAttribute(const char *key,
								  const char */*name*/)
{
//	dprintf(D_FULLDEBUG, "deleteAttribute: %s[%s]\n", key, name);

	if (!g_publish_jobs) return;

	markDirty(key);
}


MgmtScheddPlugin::JobTableType *
MgmtScheddPlugin::getJobTable()
{
	return jobTable;
}


int
MgmtScheddPlugin::HandleMgmtSocket(Service *, Stream *)
{
	singleton->getInstance()->pollCallbacks();

	return KEEP_STREAM;
}


void
MgmtScheddPlugin::processDirtyJobs()
{
	PROC_ID id;
	JobObject *job;
	ClassAd *jobAd;

	if (!g_publish_jobs) return;

	if (!scheduler) {
		scheduler = new SchedulerObject(singleton->getInstance());
	}

	for (DirtyJobsType::iterator key = dirtyJobs->begin();
		 key != dirtyJobs->end();
		 key++) {
			// Skip any key that doesn't point to an actual job
		if (!IS_JOB((*key).c_str())) {
			continue;
		}

//		dprintf(D_FULLDEBUG, "Processing: %s\n", (*key).c_str());

		id = getProcByString((*key).c_str());
		if (id.cluster < 0 || id.proc < 0) {
			dprintf(D_FULLDEBUG, "Failed to parse key: %s - skipping\n", (*key).c_str());
			continue;
		}

			// Lookup the job ad assocaited with the key. If it is not
			// present, skip the key
		if (NULL == (jobAd = GetJobAd(id.cluster, id.proc, false))) {
			dprintf(D_ALWAYS,
					"NOTICE: Failed to lookup ad for %s - maybe deleted\n",
					(*key).c_str());
			continue;
		}

			// Lookup the mgmt Job associated with the key
		if (-1 == jobTable->lookup(id, job)) {
				// If it is not present create a new one
			SubmitterObject *submitter;
			MyString name;
			if (!GetSubmitterNameFromJobAd(*jobAd, name)) {
				EXCEPT("Failed to get submitter name from job ad");
			}

			if (!GetSubmitter(name, submitter)) {
				EXCEPT("Failed to get submitter: %s", name.Value());
			}

			job = new JobObject(singleton->getInstance(),
								scheduler,
								(uint32_t) id.cluster,
								(uint32_t) id.proc,
								submitter);
			if (!job) {
				dprintf(D_ALWAYS,
						"ERROR: Failed to create job, no memory - skipping\n");
				continue;
			}

			jobTable->insert(id, job);
		}

		job->update(*jobAd);
	}

	dirtyJobs->clear();

	isHandlerRegistered = false;
}


void
MgmtScheddPlugin::markDirty(const char *key)
{
	if (!IS_JOB(key)) return;

	dirtyJobs->insert(string(key));

	if (!isHandlerRegistered) {
		daemonCore->Register_Timer(0,
								   (Eventcpp)
								   &MgmtScheddPlugin::processDirtyJobs,
								   "Process Dirty",
								   this);
		isHandlerRegistered = true;
	}
}


bool
MgmtScheddPlugin::GetSubmitter(MyString &name, SubmitterObject *&submitter)
{
	if (submitterAds->lookup(name, submitter)) {
		submitter = new SubmitterObject(singleton->getInstance(),
										scheduler,
										name.Value());
		if (!submitter) {
			dprintf(D_ALWAYS, "ERROR: Could not allocate Submitter\n");

			return false;
		}

			// Ignore old value, if it existed (returned)
		submitterAds->insert(name, submitter);
	}

	return true;
}
